/*---------------------------------------------------------------------------*/
/* p g m s o b e l   --   find edges using the Sobel operator                */
/*---------------------------------------------------------------------------*/
/* This program is written in ISO (ANSI) C.                                  */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Standard include files                                                    */
/*---------------------------------------------------------------------------*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <pbmplus.h>


/*---------------------------------------------------------------------------*/
/* Local data types                                                          */
/*---------------------------------------------------------------------------*/

typedef enum {FALSE, TRUE} logical;
typedef short int pixel_t;


/*---------------------------------------------------------------------------*/
/* Global variables                                                          */
/*---------------------------------------------------------------------------*/

char *progname = "<unknown>";


/*---------------------------------------------------------------------------*/
/* Function prototypes                                                       */
/*---------------------------------------------------------------------------*/

pixel_t **create_im (int, int);
void output_im (pixel_t **, int, int, char *);
pixel_t **read_im (FILE *, int *, int *);
void sobel (pixel_t **, int, int, pixel_t **);


/*---------------------------------------------------------------------------*/
/* M a i n   p r o g r a m                                                   */
/*---------------------------------------------------------------------------*/

int
main (int argc, char *argv[])
{
  FILE *fin;
  int m, n;
  pixel_t **im, **out;


  /* Handle the command-line argument, if any. */

  progname = argv[0];
  if (argc == 2) {
    fin = fopen (argv[1], "rb");
    if (fin == (FILE *)NULL) {
      fprintf (stderr, "Cannot open \"%s\".\n", argv[1]);
      exit (EXIT_FAILURE);
    }
  } else {
    fin = stdin;
  }

  /* Process the image. */

  im = read_im (fin, &m, &n);
  out = create_im (m, n);
  sobel (im, m, n, out);
  output_im (out, m, n, "-");
  return EXIT_SUCCESS;
}


/*---------------------------------------------------------------------------*/
/* s o b e l   --   run the Sobel edge detector over an image                */
/*---------------------------------------------------------------------------*/

void
sobel (pixel_t **im, int m, int n, pixel_t **out)
{
  int i, j;
  pixel_t v1, v2;

  /* Fill in the edges. */
  for (i = 0; i < m; i++) {
    out[0][i] = out[n-1][0] = 0;
  }

  for (j = 0; j < n; j++) {
    out[j][0] = out[j][m-1] = 0;
  }

  /* Do the main part of the image. */

  for (j = 1; j < n - 1; j++) {
    for (i = 1; i < m - 1; i++) {
      /*
      v1 = im[j+1][i-1] + 2*im[j+1][i] + im[j+1][i+1] -
	   im[j-1][i-1] - 2*im[j-1][i] - im[j-1][i+1];
      v2 = im[j-1][i-1] + 2*im[j][i-1] + im[j+1][i-1] -
	   im[j+1][i-1] + 2*im[j][i+1] - im[j+1][i+1];
      out[j][i] = (pixel_t) (sqrt ((float)(v1*v1) + (float)(v2*v2)) + 0.5);
      */
      out[j][i] = im[j-1][i-1] + im[j-1][i] + im[j-1][i+1] +
	          im[j][i-1]   - 8*im[j][i] + im[j][i+1] +
	          im[j+1][i-1] + im[j+1][i] + im[j+1][i+1];
    }
  }
}


/*---------------------------------------------------------------------------*/
/* c r e a t e _ i m   --   create an image structure                        */
/*---------------------------------------------------------------------------*/

pixel_t **
create_im (int m, int n)
{
  int i;
  pixel_t **p;

  p = (pixel_t **)malloc(n * sizeof(pixel_t *));
  if (p == (pixel_t **)NULL)
    return p;

  for (i = 0; i < n; i++) {
    p[i] = (pixel_t *)malloc(m * sizeof(pixel_t));
    if( p[i] == (pixel_t *)NULL)
      return (pixel_t **)0;
  }
  return p;
}


/*---------------------------------------------------------------------------*/
/* r e a d _ i m   --   read in an image from a PGM file                     */
/*---------------------------------------------------------------------------*/

pixel_t **
read_im (FILE *f, int *m, int *n)
{
  int i, j, format;
  gray maxval, *buf;
  pixel_t **p;

  pgm_readpgminit (f, m, n, &maxval, &format);
  p = create_im (*m, *n);

  buf = pgm_allocrow (*m);
  for (j = 0; j < *n; j++) {
    pgm_readpgmrow (f, buf, *m, maxval, format);
    for (i = 0; i < *m; i++) {
      p[j][i] = buf[i];
    }
  }
  pgm_freerow (buf);
  return p;
}


/*---------------------------------------------------------------------------*/
/* o u t p u t _ i m   --   output an image                                  */
/*---------------------------------------------------------------------------*/

void
output_im (pixel_t **p, int m, int n, char *f)
{
  FILE *fout;
  int i, j;
  gray *line;
  logical close;

  if (strcmp (f, "-")) {
    close = TRUE;
    fout = fopen (f, "wb");
    if (fout == (FILE *)NULL) {
      fprintf (stderr, "%s -- fatal: can't open %s\n", progname, f);
      exit (1);
    }
  } else {
    close = FALSE;
    fout = stdout;
  }

  /* Write the output in PBMPLUS format. */

  pgm_writepgminit (fout, m, n, (gray)255, 0);
  line = pgm_allocrow (m);
  for (j = 0; j < n; j++) {
    for (i = 0; i < m; i++)
      line[i] = (gray)p[j][i];
    pgm_writepgmrow (fout, line, m, (gray)255, 0);
  }
  pgm_freerow (line);
  if (close) {
    fclose (fout);
  }
}
